/* **********************************************************
 * Copyright (C) 1998-1999 VMWare, Inc.
 * All Rights Reserved
 * $Id: modulecall.h,v 1.13 1999/08/06 02:17:27 anil Exp $
 * **********************************************************/


/*
 * modulecall.h
 *
 *        Monitor <-->  Module (kernel driver) interface
 */

#ifndef _MODULECALL_H
#define _MODULECALL_H

#include "x86.h"

/*
 *----------------------------------------------------------------------
 *
 * ModuleCallType --
 *
 *      Enumaration of support calls done by the module
 *
 *----------------------------------------------------------------------
 */

typedef enum ModuleCallType {
   MODULECALL_NONE = 100,            //
   
   /*
    * All I/O interrupts go under MODULECALL_INTR, special-cased
    * in task switching code
    */
   MODULECALL_INTR,                  
   MODULECALL_LOCKPAGE,                  
   MODULECALL_UNLOCKPAGE,                  

   MODULECALL_ISMPNLOCKED, 
   /*
    * calls to user mode
    */
   MODULECALL_USERRETURN,            // return code for USER RPC
   MODULECALL_LAST                   // Number of entries. Must be the last one
} ModuleCallType;

typedef enum IOCTLCmd { 
   IOCTLCMD_NONE=200,
   IOCTLCMD_INIT,
   IOCTLCMD_LATE_INIT,
   IOCTLCMD_RUN,
   IOCTLCMD_LOOKUPMPN,
   IOCTLCMD_BEEP, 
   IOCTLCMD_SETUID, // Linux and VMX86_DEVEL only
   IOCTLCMD_LOCKPAGE, 
   IOCTLCMD_UNLOCKPAGE, 
   IOCTLCMD_ISMPSAFE,
   IOCTLCMD_APICBASE,
   IOCTLCMD_IOAPICBASE,
   IOCTLCMD_REGISTER_PASSTHROUGH_IO,
   IOCTLCMD_REGISTER_PASSTHROUGH_IRQ,
   IOCTLCMD_FREE_PASSTHROUGH_IO,
   IOCTLCMD_FREE_PASSTHROUGH_IRQ,
   IOCTLCMD_START_PASSTHROUGH,
   IOCTLCMD_STOP_PASSTHROUGH,
   IOCTLCMD_QUERY_PASSTHROUGH,
   IOCTLCMD_GET_NUM_VMS,
   IOCTLCMD_DECLARE_SLAVE,
   IOCTLCMD_CHECK_MEMORY,
   IOCTLCMD_BLUE_SCREEN,
   IOCTLCMD_ALLOW_CORE_DUMP,
   IOCTLCMD_GETSTATS,
   IOCTLCMD_SETSTATS,
   IOCTLCMD_SET_HARD_LIMIT,
   IOCTLCMD_GET_HARD_LIMIT,
   IOCTLCMD_GET_SET_INFO,
   IOCTLCMD_GET_INFO,
   IOCTLCMD_SET_INFO,
   IOCTLCMD_LAST
} IOCTLCmd;


#define MODULECALL_USER_START 300
#define MODULECALL_USER_END   399

#define MODULECALL_CROSS_PAGE_LEN    1
#define MODULECALL_CROSS_PAGE_START  2

// we don't want it in the final version
//#define USE_BLUE_SCREEN

#ifdef _WIN32


#define FILE_DEVICE_VMX86        0x8101
#define VMX86_IOCTL_BASE_INDEX   0x801

#define IOCTL_VMX86_INIT        CTL_CODE(FILE_DEVICE_VMX86 , \
                                 VMX86_IOCTL_BASE_INDEX+IOCTLCMD_INIT,  \
                                             METHOD_BUFFERED,     \
                                              FILE_ANY_ACCESS)

#define IOCTL_VMX86_LATE_INIT   CTL_CODE(FILE_DEVICE_VMX86 , \
                                 VMX86_IOCTL_BASE_INDEX+IOCTLCMD_LATE_INIT,  \
                                             METHOD_BUFFERED,     \
                                              FILE_ANY_ACCESS)

// We are using the IRP Information filed for the return value.
// We did it to be faster since this IOCTL is use a lot.
#define IOCTL_VMX86_RUN        CTL_CODE(FILE_DEVICE_VMX86 , \
                                 VMX86_IOCTL_BASE_INDEX+IOCTLCMD_RUN,  \
                                             METHOD_NEITHER,     \
                                              FILE_ANY_ACCESS)


#define IOCTL_VMX86_LOOKUPMPN   CTL_CODE(FILE_DEVICE_VMX86 , \
                                 VMX86_IOCTL_BASE_INDEX+IOCTLCMD_LOOKUPMPN,  \
                                             METHOD_BUFFERED,     \
                                              FILE_ANY_ACCESS)

#define IOCTL_VMX86_BEEP   CTL_CODE(FILE_DEVICE_VMX86 , \
                                 VMX86_IOCTL_BASE_INDEX+IOCTLCMD_BEEP,  \
                                             METHOD_BUFFERED,     \
                                              FILE_ANY_ACCESS)

#define IOCTL_VMX86_LOCKPAGE   CTL_CODE(FILE_DEVICE_VMX86 , \
                                 VMX86_IOCTL_BASE_INDEX+IOCTLCMD_LOCKPAGE,  \
                                             METHOD_BUFFERED,     \
                                              FILE_ANY_ACCESS)

#define IOCTL_VMX86_UNLOCKPAGE   CTL_CODE(FILE_DEVICE_VMX86 , \
                                 VMX86_IOCTL_BASE_INDEX+IOCTLCMD_UNLOCKPAGE,  \
                                             METHOD_BUFFERED,     \
                                              FILE_ANY_ACCESS)

#define IOCTL_VMX86_APICBASE   CTL_CODE(FILE_DEVICE_VMX86 , \
                                 VMX86_IOCTL_BASE_INDEX+IOCTLCMD_APICBASE,  \
                                             METHOD_BUFFERED,     \
                                              FILE_ANY_ACCESS)

#define IOCTL_VMX86_IOAPICBASE   CTL_CODE(FILE_DEVICE_VMX86 , \
                                 VMX86_IOCTL_BASE_INDEX+IOCTLCMD_IOAPICBASE,  \
                                             METHOD_BUFFERED,     \
                                              FILE_ANY_ACCESS)

#ifdef USE_BLUE_SCREEN
#define IOCTL_VMX86_BLUE_SCREEN  CTL_CODE(FILE_DEVICE_VMX86 , \
                                 VMX86_IOCTL_BASE_INDEX+IOCTLCMD_BLUE_SCREEN,  \
                                             METHOD_BUFFERED,     \
                                              FILE_ANY_ACCESS)
#endif

#define IOCTL_VMX86_GETSTATS	 CTL_CODE(FILE_DEVICE_VMX86 , \
                                 VMX86_IOCTL_BASE_INDEX+IOCTLCMD_GETSTATS,  \
                                             METHOD_BUFFERED,     \
                                              FILE_ANY_ACCESS)

#define IOCTL_VMX86_SETSTATS	 CTL_CODE(FILE_DEVICE_VMX86 , \
                                 VMX86_IOCTL_BASE_INDEX+IOCTLCMD_SETSTATS,  \
                                             METHOD_BUFFERED,     \
                                              FILE_ANY_ACCESS)

#define IOCTL_VMX86_SETHARDLIMIT	 CTL_CODE(FILE_DEVICE_VMX86 , \
                                       VMX86_IOCTL_BASE_INDEX+IOCTLCMD_SET_HARD_LIMIT,  \
                                             METHOD_BUFFERED,     \
                                              FILE_ANY_ACCESS)

#define IOCTL_VMX86_GETHARDLIMIT	 CTL_CODE(FILE_DEVICE_VMX86 , \
                                          VMX86_IOCTL_BASE_INDEX+IOCTLCMD_GET_HARD_LIMIT,  \
                                             METHOD_BUFFERED,     \
                                              FILE_ANY_ACCESS)

#define IOCTL_VMX86_GETSETINFO	         CTL_CODE(FILE_DEVICE_VMX86 , \
                                          VMX86_IOCTL_BASE_INDEX+IOCTLCMD_GET_SET_INFO,  \
                                             METHOD_BUFFERED,     \
                                              FILE_ANY_ACCESS)

#define IOCTL_VMX86_GETINFO	         CTL_CODE(FILE_DEVICE_VMX86 , \
                                          VMX86_IOCTL_BASE_INDEX+IOCTLCMD_GET_INFO,  \
                                             METHOD_BUFFERED,     \
                                              FILE_ANY_ACCESS)

#define IOCTL_VMX86_SETINFO	         CTL_CODE(FILE_DEVICE_VMX86 , \
                                          VMX86_IOCTL_BASE_INDEX+IOCTLCMD_SET_INFO,  \
                                             METHOD_BUFFERED,     \
                                              FILE_ANY_ACCESS)

#define IOCTL_VMX86_GETNUMVMS	         CTL_CODE(FILE_DEVICE_VMX86 , \
                                          VMX86_IOCTL_BASE_INDEX+IOCTLCMD_GET_NUM_VMS,  \
                                             METHOD_BUFFERED,     \
                                              FILE_ANY_ACCESS)

#endif

/*
 *----------------------------------------------------------------------
 *
 * VMCrossPage --
 *
 *      data structure shared between the monitor and the module
 *      that is used for crossing between the two.
 *      Accessible as vm->cross (kernel module) and CROSS_PAGE
 *      (monitor)
 *
 *      Exactly one page long
 *
 *----------------------------------------------------------------------
 */

#define CONTEXT_SWITCH_CODE_SIZE 192
#define MAX_PASSTHROUGH_PORT 1024
#define NUM_PASSTHROUGH_WORDS (MAX_PASSTHROUGH_PORT / 32)

#ifdef __GNUC__
 struct ContextInfo {
#else
 typedef  struct ContextInfo {
#endif
    DTRWords    gdtr;
    DTRWords    ldtr;
    DTRWords    idtr;
    uint16 tr;
    uint8 jump[6];
    Task task;   
#ifdef __GNUC__
 } __attribute__ ((packed));

typedef struct ContextInfo ContextInfo;
#else
 } ContextInfo;
#endif

#ifdef __GNUC__
 struct VMCrossPage {
#else
 typedef struct VMCrossPage {
#endif
   /*
    * Version checking. Should remain at offset 0
    */ 
   uint32 version; 
  

   /*
    * Changes in contextSwitchCode location need to
    * be reflected in addrlayout.h:COMINGBACK_OFFSET
    * (ASSERT fail otherwise)
    */
   uint8 contextSwitchCode[CONTEXT_SWITCH_CODE_SIZE];

   /*
    * iobitMap must immediately follow monContext.task
    * (ASSERT fail otherwise)
    */
   ContextInfo hostContext;
   ContextInfo monContext;
   uint32      iobitMap[NUM_PASSTHROUGH_WORDS];

   struct VMCrossPage * hostVA;
   PTE    hostSharedPTE;

   uint32 hostContextVA;


   uint32 switchPDOffset;     // index into the page directory of switchPDE
   PDE  switchPDE;            // PDE that points to switch page table


   LA monitorLinearAddr;      // base location of monitor in linear space
   LA crosspageLinearAddr;    // location of crosspage page in both linear spaces

   //
   // ModuleCall parameters
   //
   ModuleCallType moduleCallType;
   uint32 args[4];
   uint32 retval;

   VA userSharedAreaAddr;

   // host irq relocation values
   int irqRelocateOffset[2]; 

#if defined(WINNT_DDK) && defined(VMX86_DEBUG)
   void *vmPtr;  /* Monitor to VmDriver structure, useful for debugging */
#endif

#ifdef __GNUC__
 } __attribute__ ((packed));

typedef struct VMCrossPage VMCrossPage;
#else
 } VMCrossPage;
#endif

#define MODULECALL_CROSSPAGE_SIZE 0x294
#define MODULECALL_CROSSPAGE_LAOFFSET 0x26c


#define CROSSPAGE_VERSION (0x1776 + VERSION_MAGIC)


#define MONITOR_LINEAR_START            (CROSS_PAGE->monitorLinearAddr)
   
#ifdef MONITOR
#define CROSS_PAGE  ((VMCrossPage *)(MONITOR_BASE_VADDR + VPN_2_VA(CROSS_PAGE_START)))
#endif

struct passthrough_iorange {
  unsigned short ioBase;	/* Base of range to pass through */
  unsigned short numPorts;	/* Length of range */
};

typedef struct {
  uint32	userLockedPages;	/* current count of pages locked by app */
  uint32	maxUserLockedPages;	/* max over time of pages locked by app */
  uint32	monitorLockedPages;	/* current count of pages locked by monitor */
  uint32	maxMonitorLockedPages;	/* max over time of pages locked by monitor */
} VMX86Stats;

/*
 * Return codes from lock page if lock page failed.
 *
 *    PAGE_LOCK_FAILED              The host refused to lock a page.
 *    PAGE_LOCK_LIMIT_EXCEEDED      We have reached the limit of locked pages for all VMs
 */
#define PAGE_LOCK_FAILED                  0
#define PAGE_LOCK_LIMIT_EXCEEDED         -1

#endif
