/* **********************************************************
 * Copyright (C) 1998-1999 VMWare, Inc.
 * All Rights Reserved
 * $Id: vm_time.h,v 1.4.8.1 1999/08/27 00:18:05 bhlim Exp $
 * **********************************************************/


/*
 * vm_time.h  --
 *
 *    Time management functions.
 *    Part of driver-only distribution
 *
 *    see comment in poll.c
 */


#ifndef VM_TIME_H
#define VM_TIME_H


#include "vm_types.h"

typedef int64 VmTimeType;          /* Time in microseconds */
typedef int64 VmTimeRealClock;     /* Real clock kept in microseconds */
typedef int64 VmTimeVirtualClock;  /* Virtual Clock kept in CPU cycles */



typedef struct VmTimeStart {
   VmTimeVirtualClock   count;          
   VmTimeRealClock      time; 
   uint32		mhzEstimate;    // Estimate of MHZ cycle time
   uint32		pad;		// pad to 64-bit boundary
} VmTimeStart;


typedef struct VmTimeVirtualRealClock {
   VmTimeType		lastVirtualTime;
   VmTimeType		lastRealTime;
   VmTimeType		lastVirtualRealTime;
   VmTimeType		error;			// accumulated error
   double		maxCorrectionRate;	// ratio of times
   struct VM		*vm;			// backpointer to vm
} VmTimeVirtualRealClock;

#define VMTIME_VIRTUAL_INFINITE  ((((VmTimeVirtualClock)1)<<62)-1)

#define CYCLES_TO_USECS(_c)   ((_c) / sharedArea->startTime.mhzEstimate)
#define USECS_TO_CYCLES(_us)  ((_us) * sharedArea->startTime.mhzEstimate)

#define GET_REALCYCLES()      (GET_TSC())
#define GET_REALTIME()        (GET_TSC() / sharedArea->startTime.mhzEstimate)

#define GET_VIRTUALCYCLES()   (sharedArea->vcpu[0].currentVirtualTime)
#define GET_VIRTUALCYCLES32() (*(uint32 *)&(sharedArea->vcpu[0].currentVirtualTime))
#define GET_VIRTUALTIME() \
   (sharedArea->vcpu[0].currentVirtualTime/sharedArea->startTime.mhzEstimate)

#endif /* VM_TIME_H */

