/* **********************************************************
 * Copyright (C) 1998 VMWare, Inc.
 * All Rights Reserved
 * **********************************************************/
#ifdef VMX86_DEVEL
char rcsId_driver[] = "$Id: driver.c,v 1.21 1999/08/06 22:39:55 chris Exp $";
#else
#define FILECODE "F(411)"
#endif 

#include "driver-config.h"

#ifdef KERNEL_2_1
#  define EXPORT_SYMTAB
#  ifndef __KERNEL__
#    define __KERNEL__
#  endif
#  define memcpy_fromfs	copy_from_user
#  define memcpy_tofs	copy_to_user
#  define KFREE_SKB(skb, type)		kfree_skb(skb)
#  define DEV_KFREE_SKB(skb, type)	dev_kfree_skb(skb)
#  define SKB_INCREF(skb)		atomic_inc(&(skb)->users)
#  define SKB_IS_CLONE_OF(clone, skb)	( \
      skb_datarefp(clone) == skb_datarefp(skb) \
   )
#  define SK_ALLOC(pri)			sk_alloc(0, pri, 1)
#  define DEV_QUEUE_XMIT(skb, dev, pri)	( \
      (skb)->dev = (dev), \
      (skb)->priority = (pri), \
      (skb)->mac.raw = (skb)->data, \
      (skb)->nh.raw = (skb)->data + sizeof (struct ethhdr), \
      dev_queue_xmit(skb) \
   )
#  define DEV_LOCK_LIST()		dev_lock_list()
#  define DEV_UNLOCK_LIST()		dev_unlock_list()
#else
#  define KFREE_SKB(skb, type)		kfree_skb(skb, type)
#  define DEV_KFREE_SKB(skb, type)	dev_kfree_skb(skb, type)
#  define SKB_INCREF(skb)		atomic_inc(&(skb)->count)
#  define SKB_IS_CLONE_OF(clone, skb)	( \
      (clone)->data_skb != NULL && \
      ((clone)->data_skb == (skb) || (clone)->data_skb == (skb)->data_skb) \
   )
#  define SK_ALLOC(pri)			sk_alloc(pri)
#  define DEV_QUEUE_XMIT(skb, dev, pri)	dev_queue_xmit(skb, dev, pri)
#  define DEV_LOCK_LIST()		// dev_lockct missing from ksyms
#  define DEV_UNLOCK_LIST()		// dev_lockct missing from ksyms
#endif

#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/sched.h>
#include <linux/malloc.h>
#ifdef KERNEL_2_1
#include <linux/poll.h>
#include <asm/uaccess.h>
#endif

#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/mm.h>
#include <linux/skbuff.h>
#include <linux/if_ether.h>
#include <linux/sockios.h>
#include <net/sock.h>

#define __KERNEL_SYSCALLS__
#include <asm/io.h>

#include "vm_oui.h"

#include <linux/proc_fs.h>
#include <linux/file.h>

/* 
 * Logging
 */

#ifdef VMX86_RELEASE
#define LOGLEVEL 0
#else
#define LOGLEVEL 1
#endif

#if LOGLEVEL >= 0
#define LOG(level, args) ((void) (LOGLEVEL >= (level) ? (printk args) : 0))
#else
#define LOG(level, args)
#endif

#if 0
static unsigned char watchAddr[6] = { 0x00, 0x50, 0x56, 0x82, 0x45, 0xa4 };
#define WATCH(skb) ( \
      MAC_EQ(SKB_2_DESTMAC(skb), watchAddr) || \
      MAC_EQ(SKB_2_SRCMAC(skb), watchAddr) \
   )
#else
#define WATCH(skb) FALSE
#endif


/*
 * Fundamental sizes
 */

#define VNET_NUM_VNETS         256
#define VNET_MAJOR_NUMBER      119

#define VNET_PORTS_PER_HUB     9
#define VNET_NUM_HUBS          VNET_NUM_VNETS
#define VNET_MAX_QLEN          32

#define VNET_BRIDGE_HISTORY    8

#define VNET_NUM_IPBASED_MACS  64  // there can only be 63 IP-based MAC addresses,
                                   // numbered from 1 to 63; the rest are random.



/*
 * Ethernet
 */

#define MAC_EQ(_a, _b)         !memcmp((_a), (_b), ETH_ALEN)
#define SKB_2_DESTMAC(_skb)    (((struct ethhdr *)(_skb)->data)->h_dest)
#define SKB_2_SRCMAC(_skb)     (((struct ethhdr *)(_skb)->data)->h_source)
#define UP_AND_RUNNING(_flags) (((_flags) & (IFF_RUNNING|IFF_UP)) == \
				(IFF_RUNNING|IFF_UP))


/*
 * Types
 */

typedef u_char uint8;
typedef u_short uint16;
typedef u_long uint32;
typedef long int32;
typedef char Bool;

#define TRUE  1
#define FALSE 0


/*
 * Data structures
 */

typedef struct VNetHub VNetHub;
typedef struct VNetPort VNetPort;

typedef struct VNetBridge {
   struct notifier_block    notifier;	// must be first
   char                     name[8];
   struct device           *dev;
   struct sock             *sk;
   struct packet_type       pt;
   VNetPort                *port;
   Bool                     promiscSaved;
   Bool                     savedPromisc;
   struct sk_buff          *history[VNET_BRIDGE_HISTORY];
} VNetBridge;

struct VNetPort {
   uint32                   flags;
   uint8                    paddr[ETH_ALEN];
   uint8                    ladrf[8]; // 64-bit mask
   struct sk_buff_head      packetQueue;
   VNetHub                 *hub;
#ifndef KERNEL_2_3_1
   struct wait_queue       *waitQueue;
#else
   wait_queue_head_t        waitQueue;
#endif
   VNetBridge              *bridge;
};

struct VNetHub {
   VNetPort                 port[VNET_PORTS_PER_HUB];
   int                      portCount;
   int                      bridgeCount;
   struct device           *netIF;
   char                     devName[8];
   struct enet_statistics   stats;
   Bool                     keep;  // this is off by default
};

/* a pointer table to the actual hubs. */
static VNetHub *vnetHub[VNET_NUM_HUBS];

#if LOGLEVEL >= 4
static struct timeval vnetTime;
#endif

/* this will let all multicast packets go through. */
static uint8 AllMultiFilter[8] = { 0xff, 0xff, 0xff, 0xff,
				   0xff, 0xff, 0xff, 0xff };

/*
 *  Device driver interface.
 */

static int  VNetFileOpOpen(struct inode *inode, struct file *filp);
#ifdef KERNEL_2_1
static int  VNetFileOpClose(struct inode *inode, struct file *filp);
static unsigned int VNetFileOpPoll(struct file *filp, poll_table *wait);
static int  VNetFileOpRead(struct file *filp, char *buf, size_t count,
			   loff_t *ppos);
static int  VNetFileOpWrite(struct file *filp, const char *buf, size_t count,
			    loff_t *ppos);
#else
static void VNetFileOpClose(struct inode *inode, struct file *filp);
static int  VNetFileOpSelect(struct inode *inode, struct file *filp,
                             int sel_type, select_table *wait);
static int  VNetFileOpRead(struct inode *inode, struct file *filp,
                            char *buf, int count);
static int  VNetFileOpWrite(struct inode *inode, struct file *filp,
                             const char *buf, int count);
#endif
static int  VNetFileOpIoctl(struct inode *inode, struct file *filp,
                            unsigned int iocmd, unsigned long ioarg);

#ifdef KERNEL_2_1
static struct file_operations vnetFileOps = {
	NULL,              /* lseek */
	VNetFileOpRead,    /* read */
	VNetFileOpWrite,   /* write */
	NULL,              /* readdir */
	VNetFileOpPoll,    /* select */
	VNetFileOpIoctl,   /* ioctl */
	NULL,              /* mmap */
	VNetFileOpOpen,    /* open */
	NULL,		   /* flush */
	VNetFileOpClose,   /* release */
	NULL               /* fsync */
};
#else
static struct file_operations vnetFileOps = {
	NULL,              /* lseek */
	VNetFileOpRead,    /* read */
	VNetFileOpWrite,   /* write */
	NULL,              /* readdir */
	VNetFileOpSelect,  /* select */
	VNetFileOpIoctl,   /* ioctl */
	NULL,              /* mmap */
	VNetFileOpOpen,    /* open */
	VNetFileOpClose,   /* release */
	NULL               /* fsync */
};
#endif


/*
 *  netif interface
 */

static int  VNetNetifOpen(struct device *dev);
static int  VNetNetifProbe(struct device *dev);
static int  VNetNetifClose(struct device *dev);
static int  VNetNetifStartXmit(struct sk_buff *skb, struct device *dev);
static struct enet_statistics *VNetNetifGetStats(struct device *dev);
static int  VNetNetifSetMAC(struct device *dev, void *addr);
static void VNetNetifSetMulticast(struct device *dev);

/*
 * Bridging
 */

static int  VNetBridgeAttach(VNetPort *port, unsigned long ioarg);
static void VNetBridgeDetach(VNetPort *port);
static int  VNetBridgeUp(VNetPort *port);
static void VNetBridgeDown(VNetPort *port);
static int  VNetBridgeNotify(struct notifier_block *this, u_long msg,
			     void *data);
static int  VNetBridgeReceive(struct sk_buff *skb, struct device *dev,
			      struct packet_type *pt);
static void VNetBridgeXmitPacket(VNetPort *port, struct sk_buff *skb);
static void VNetBridgeStartPromisc(VNetPort *port);
static void VNetBridgeStopPromisc(VNetPort *port);
static void VNetBridgeCheckPromisc(VNetPort *port);


/*
 * Utility functions
 */

static void VNetXmitPacket(VNetHub *hub, VNetPort *src, struct sk_buff *skb);
static Bool VNetPacketMatch(uint8 *destAddr, uint8 *ifAddr, uint8 *ladrf, uint32 flags);
static void VNetMakeRandomAddress(uint8 *addr);
static int VNetOpenHub(int vnetHub, Bool keep);

#ifdef EDWARD
static void VNetPacketDebug(struct sk_buff *skb);
static uint16 VNetICMPChecksum(uint16 *packet, int len);
#endif

/*
 * procfs interface (/proc/vmnet)
 */
static int VNetProcFileRead(char *buf, char **start, off_t pos, int count, int wr);
struct proc_dir_entry VNetProcFile = 
{
	0,           // inode number -- dynamic, ignore
	5,           // length of "vmnet"
	"vmnet",     // filename */
	S_IFREG | S_IRUGO, /* File mode - this is a regular 
                            * file which can be read by its 
			    * owner, its group, and everybody
			    * else */
	1,           // number of links
	0, 0,        // uid, gid 
	0,           // filesize in "ls"
	NULL,        /* functions which can be done on the inode 
	              * (linking, removing, etc.) - we don't 
		      * support any. */
	VNetProcFileRead, // read function for this file 
	NULL          // XXX 2.0.x had extra function, not in 2.2?
}; 

/*
 *----------------------------------------------------------------------
 *
 * VNetRegister --
 *
 *      (debugging support) Should be the first function of this file
 *
 * Results:
 *      
 *      Registers the module. 
 *      /sbin/ksyms -a | grep VNetRegister will return the base
 *      address of that function as loaded in the kernel. 
 * 
 *      Since this is the first function of the kernel module,
 *      every other symbol can be computing by adding the base
 *      to the output of nm.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

int VNetRegister(int);

#ifdef KERNEL_2_1
EXPORT_SYMBOL(VNetRegister);
#else
static struct symbol_table vnet_syms = {
#include <linux/symtab_begin.h>
   X(VNetRegister),
#include <linux/symtab_end.h>
};
#endif

int 
VNetRegister(int value)
{
   LOG(0, (KERN_WARNING "/dev/vmnet: VNetRegister called\n"));
   return 0;
}


/*
 *----------------------------------------------------------------------
 *
 * init_module --
 *
 *      linux module entry point. Called by /sbin/insmod command
 * 
 * Results: 
 *      registers a device driver for a major # that depends
 *      on the uid. Add yourself to that list.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

int 
init_module(void)
{
   int retval;
   int i;

#ifdef KERNEL_2_1
   retval = proc_register(&proc_root, &VNetProcFile);
#else // proc_register_dynamic disappeared as of 2.1
   retval = proc_register_dynamic(&proc_root, &VNetProcFile);
#endif
   if (retval) {
     LOG(0, (KERN_NOTICE "/dev/vmnet: could not register /proc/vmnet\n"));
     return -EINVAL;
   }

   for (i=0; i<VNET_NUM_HUBS; i++) {   // initialize pointer array
     vnetHub[i] = NULL;
   }
   
   retval = register_chrdev(VNET_MAJOR_NUMBER, "vmnet", &vnetFileOps);
   if (retval) {
      LOG(0, (KERN_NOTICE "/dev/vmnet: could not register major device %d\n",
	      VNET_MAJOR_NUMBER));
      return -ENOENT;
   }

   /* 
    * allow vmnet1 to be set up by default.  this is because VMware
    * is shipped using vmnet1 for pre-configured, hostonly networking,
    * and our rc.d startup scripts expect vmnet1 to be an existing
    * interface at boot time. "keep" flag should is set to TRUE so we 
    * don't lose the networking configuration.
    *
    * the vmnet-bridge program opens vmnet0 for itself, so it's okay.
    */
   retval = VNetOpenHub(1, TRUE);
   if (retval) {
      LOG(0, (KERN_NOTICE "/dev/vmnet: could not register vmnet1\n"));
      return -ENOENT;
   }

#ifndef KERNEL_2_1      
   register_symtab(&vnet_syms);
#endif

#ifdef LINUX_VMWARE
   VMWare_SDumpInit();
   VMWare_SDumpSend("Networking without panics\n");
#endif
   return 0;
}


/*
 *----------------------------------------------------------------------
 *
 * cleanup_module --
 *
 *      Called by /sbin/rmmod
 *
 * Results: 
 *      None.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

int
cleanup_module(void)
{
   int retval;
   int i, j;

   for (i=0; i<VNET_NUM_HUBS; i++) {
      if (!vnetHub[i]) {   // don't free null hubs
	 continue;
      }
      unregister_netdev(vnetHub[i]->netIF);
      kfree(vnetHub[i]->netIF);
      
      for (j=0; j<VNET_PORTS_PER_HUB; j++) {
         if (vnetHub[i]->port[j].flags & IFF_RUNNING) {
            LOG(0, (KERN_WARNING "%s: port %d still open on module cleanup\n",
		    vnetHub[i]->devName, j));
            // return -EBUSY;
         }
      }
      kfree(vnetHub[i]); // free remaining hubs
   }
   
   retval =  unregister_chrdev(VNET_MAJOR_NUMBER, "vmnet");
   if (retval != 0 ) {
      LOG(0, (KERN_WARNING "/dev/vmnet: could not unregister major device %d\n",
	      VNET_MAJOR_NUMBER));
   }
   
   proc_unregister(&proc_root, VNetProcFile.low_ino);
   
   return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * VNetOpenHub --
 *
 *      Usually called by VNetFileOpOpen.  Allocates memory and creates
 *      a new hub /dev/vmnetN and its corresponding interface.  Useful
 *      for both dynamically creating hubs as well as defining default
 *      hubs that should be set up at insmod-time (e.g. vmnet1).
 *
 *      Does *not* open ports; VNetFileOpOpen opens ports.
 *
 * Results: 
 *      None.
 *
 * Side effects:
 *      Hub is allocated, device is made.
 *
 *----------------------------------------------------------------------
 */

int
VNetOpenHub(int hubNum, Bool keep)
{
   VNetHub *hub;
   int j;
   
   LOG(1, (KERN_DEBUG "/dev/vmnet: opening hub %d (keep %d)\n", hubNum, keep));
   
   /*
    * check if hub exists.  if not, alloc one and assign pointer.
    */
   if (!vnetHub[hubNum]) {
      struct device *dev;
      LOG(1, (KERN_DEBUG "/dev/vmnet: hub %d does not exist, allocating memory.\n", hubNum));
      vnetHub[hubNum] = (struct VNetHub *)kmalloc(sizeof(struct VNetHub), GFP_KERNEL);
      memset(vnetHub[hubNum], 0, sizeof(struct VNetHub));
      
      dev = (struct device *)kmalloc(sizeof(struct device), GFP_KERNEL);
      memset(dev, 0, sizeof(struct device));
      dev->priv = vnetHub[hubNum];
      dev->init = VNetNetifProbe;
      sprintf(vnetHub[hubNum]->devName, "vmnet%d", hubNum);
      dev->name = vnetHub[hubNum]->devName;
      
      ether_setup(dev); // turns on IFF_BROADCAST, IFF_MULTICAST
      if (register_netdev(dev) != 0) {
	 LOG(0, (KERN_NOTICE "%s: could not register network device\n",
		 vnetHub[hubNum]->devName));
	 kfree(dev);
	 //	 return 0;
	 return -ENODEV;
      }
      
      /*
       * The initial address assignment for the netIF and ports
       * is not unique across machines.
       *
       * The last two bytes of the address can be modified later
       * to something more likely to be unique.
       */
      
      dev->dev_addr[0] = VMX86_OUI0;
      dev->dev_addr[1] = VMX86_OUI1;
      dev->dev_addr[2] = VMX86_OUI2;
      dev->dev_addr[3] = VMX86_MAC_IPBASED | hubNum * (VNET_PORTS_PER_HUB + 1);
      
      dev->open = &VNetNetifOpen;
      dev->hard_start_xmit = &VNetNetifStartXmit;
      dev->stop = &VNetNetifClose;
      dev->get_stats = &VNetNetifGetStats;
      dev->set_mac_address = &VNetNetifSetMAC;
      dev->set_multicast_list = &VNetNetifSetMulticast;
      
      vnetHub[hubNum]->netIF = dev;
      
      for (j=0; j<VNET_PORTS_PER_HUB; j++) {
	 
	 vnetHub[hubNum]->port[j].flags = 0;
	 skb_queue_head_init(&(vnetHub[hubNum]->port[j].packetQueue));
	 vnetHub[hubNum]->port[j].hub = vnetHub[hubNum];
#ifndef KERNEL_2_3_1
	 vnetHub[hubNum]->port[j].waitQueue = NULL;
#else
	 init_waitqueue_head(&vnetHub[hubNum]->port[j].waitQueue);
#endif
      }

      hub = (struct VNetHub *)vnetHub[hubNum]; // get pointer from array
      hub->keep = keep; // only change keep if hub is being created
   }
   
   return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * VNetFileOpOpen --
 *
 *      The virtual network's open file operation. Open a connection to
 *      this virtual network. (Like plugging a cable into this virtual
 *      hub.)
 *
 * Results: 
 *      Errno.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

int
VNetFileOpOpen(struct inode  *inode,
               struct file   *filp)
{
   VNetPort *port = NULL;
   VNetHub *hub;
   int hubNum, retval, i;
   
   LOG(1, (KERN_DEBUG "/dev/vmnet: open called by PID %d (%s)\n", current->pid, current->comm));

   hubNum = MINOR(inode->i_rdev);
   if ((hubNum < 0) || (hubNum >= VNET_NUM_HUBS)) {
      return -ENODEV;
   }

   // VNetOpenHub does our "already-allocated?" checking for us.
   retval = VNetOpenHub(hubNum, FALSE);  // by default new hubs aren't kept
   if (retval) {
      LOG(0, (KERN_NOTICE "/dev/vmnet: could not open hub vmnet%d\n", hubNum));
      return retval;
   }
   hub = (struct VNetHub *)vnetHub[hubNum];

   /*
    * Find a free port and take it.
    *
    * XXX there's a race with VNetFileOpClose?
    */

   for (i = 0; i < VNET_PORTS_PER_HUB; i++) {
      if ((hub->port[i].flags & IFF_RUNNING) == 0) {
         port = &hub->port[i];
         break;
      }
   }
   if (!port) {
      return -EBUSY;
   }
   port->flags = IFF_RUNNING;

   /*
    * Initialize fields
    */

   memset(port->ladrf, 0, sizeof port->ladrf);

   port->paddr[0] = VMX86_OUI0;
   port->paddr[1] = VMX86_OUI1;
   port->paddr[2] = VMX86_OUI2;

   /*
    * When first opened, all MAC addresses are randomly chosen.
    * They are subsequently set by vlance or whomever through ioctl calls.
    */

   VNetMakeRandomAddress((uint8 *)&port->paddr);

   port->bridge = NULL;
   filp->private_data = port;
   
   /*
    * Update bridging state
    */
   
   hub->portCount++;
   if (hub->portCount == 2 && hub->bridgeCount > 0) {
      for (i = 0; i < VNET_PORTS_PER_HUB; i++) {
	 VNetPort *p = &hub->port[i];
	 if (p->bridge != NULL && p->bridge->dev != NULL) {
	    VNetBridgeStartPromisc(p);
	 }
      }
   }

#ifdef LINUX_VMWARE
   {
      char buf[32];
      sprintf(buf,"VNetFileOpWrite =0x%p \n",VNetFileOpWrite);
      VMWare_SDumpInit();
      VMWare_SDumpSend(buf);
   }
#endif

   LOG(1, (KERN_DEBUG "/dev/vmnet: port %d on hub %d successfully opened\n", i, hubNum));
  
   MOD_INC_USE_COUNT;
   return 0;
}


/*
 *----------------------------------------------------------------------
 *
 * VNetFileOpClose --
 *
 *      The virtual network's close file operation. Unplugging the cable.
 *
 * Results: 
 *      None.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

#ifdef KERNEL_2_1
static int
#else
static void
#endif
VNetFileOpClose(struct inode  *inode,
                struct file   *filp)
{
   VNetPort *port = (VNetPort*)filp->private_data;
   VNetHub *hub = port->hub;
   struct sk_buff *skb;
   int i;

   /*
    * Clear port->flags first so nobody will try to give us packets.
    *
    * XXX race with VNetFileOpOpen?
    */
   
   port->flags = 0;

   /*
    * Deallocate resources
    */

   if (port->bridge != NULL) {
      VNetBridgeDetach(port);
   }

   for (;;) {
      skb = skb_dequeue(&port->packetQueue);
      if (skb == NULL) {
	 break;
      }
      DEV_KFREE_SKB(skb, FREE_WRITE);
   }

   /*
    * Update bridging state
    */

   hub->portCount--;
   if (hub->portCount == 1 && hub->bridgeCount > 0) {
      for (i = 0; i < VNET_PORTS_PER_HUB; i++) {
	 VNetPort *p = &hub->port[i];
	 if (p->bridge != NULL && p->bridge->dev != NULL) {
	    VNetBridgeStopPromisc(p);
	 }
      }
   }

   /*
    * If no more ports left on this hub, and hub->keep is false, 
    * dealloc hub's memory and pointer in *vnetHub[], and 
    * dealloc and close its iface.
    */
   if (hub->portCount == 0) {
      for (i = 0; i < VNET_NUM_HUBS; i++) {
	 if (vnetHub[i] == hub) {
	    break;
	 }
      }
      
      if (!vnetHub[i]->keep) {
	 unregister_netdev(vnetHub[i]->netIF);
	 kfree(vnetHub[i]->netIF);
	 
	 vnetHub[i] = NULL;
	 kfree(hub);
	 LOG(1, (KERN_DEBUG "/dev/vmnet: closed and deallocated hub %d\n", i));
      } else {
	 LOG(1, (KERN_DEBUG "/dev/vmnet: keep flag is %d, keeping hub %d\n", vnetHub[i]->keep, i));
      }
   }
   
   MOD_DEC_USE_COUNT;

#ifdef KERNEL_2_1
   return 0;
#endif
}


/*
 *----------------------------------------------------------------------
 *
 * VNetFileOpRead --
 *
 *      The virtual network's read file operation. Reads the next pending
 *      packet for this network connection.
 *
 * Results: 
 *      On success the len of the packet received,
 *      else if no packet waiting and nonblocking 0,
 *      else -errno.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

#ifdef KERNEL_2_1
int 
VNetFileOpRead(struct file  *filp,
               char         *buf,
               size_t        count,
	       loff_t	    *ppos)
#else
int
VNetFileOpRead(struct inode *inode,
               struct file  *filp,
               char         *buf,
               int           count)
#endif
{
   VNetPort *port = (VNetPort*)filp->private_data;
   struct sk_buff *skb;
#ifdef KERNEL_2_1
   int i;
   sigset_t sigs;
#endif   
   
   for (;;) {
      start_bh_atomic();
      skb = skb_dequeue(&port->packetQueue);
      end_bh_atomic();
      if (skb != NULL || filp->f_flags & O_NONBLOCK) {
	 break;
      }
      interruptible_sleep_on(&port->waitQueue);
#ifdef KERNEL_2_1
      signandsets(&sigs, &current->signal, &current->blocked);
      for (i = 0; i < _NSIG_WORDS; i++) {
	 if (sigs.sig[i] != 0) {
	    return -EINTR;
	 }
      }
#else
      if (current->signal & ~current->blocked) {
	 return -EINTR;
      }
#endif
   }

   if (skb) {
      count = (count < skb->len) ? count : skb->len;
      memcpy_tofs(buf, skb->data, count);
   } else {
      return -EAGAIN;
   }

   DEV_KFREE_SKB(skb, FREE_WRITE);
   return count;
}


/*
 *----------------------------------------------------------------------
 *
 * VNetFileOpWrite --
 *
 *      The virtual network's write file operation. Send the raw packet
 *      to the network.
 *
 * Results: 
 *      On success the count of bytes written else errno.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */
#ifdef KERNEL_2_1
int
VNetFileOpWrite(struct file   *filp,
                const char    *buf,
                size_t         count,
		loff_t	      *ppos)
#else
int
VNetFileOpWrite(struct inode  *inode,
                struct file   *filp,
                const char    *buf,
                int            count)
#endif
{
   VNetPort *port = (VNetPort*)filp->private_data;
   VNetHub *hub = port->hub;
   struct sk_buff *skb;

   /*
    * Check size
    */

   if (count < sizeof (struct ethhdr)) {
      return -EINVAL;
   }

   /*
    * Allocate an sk_buff.
    */

   skb = dev_alloc_skb(count + 7);
   if (skb == NULL) {
      // XXX obey O_NONBLOCK?
      return -ENOBUFS;
   }
#ifndef KERNEL_2_1
   skb->free = 1;
#endif

   skb_reserve(skb, 2);
   skb->dev = hub->netIF;

#ifdef LINUX_VMWARE
   if (intr_count) { 
      char buf[64];
      sprintf(buf,"VNetFileOpWrite intr_count=%d\n",intr_count);
      VMWare_SDumpInit();
      VMWare_SDumpSend(buf);
   }
   if (skb->tail + count > skb->end) { 
      char buf[64];
      sprintf(buf,"VNetFileOpWrite skb->tail 0x%p skb->end 0x%p len 0x%x ",
              skb->tail,skb->end,count);
      VMWare_SDumpInit();
      VMWare_SDumpSend(buf);
   }
#endif

   /*
    * Copy the data and send it.
    */

   memcpy_fromfs(skb_put(skb, count), buf, count);
   VNetXmitPacket(hub, port, skb);

   return count;
}


/*
 *----------------------------------------------------------------------
 *
 * VNetFileOpSelect --
 *
 *      The virtual network's file select operation.
 *
 * Results: 
 *      Return 1 if success, else sleep and return 0.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */
#ifdef KERNEL_2_1

static unsigned int
VNetFileOpPoll(struct file *filp,
	       poll_table *wait)
{
   VNetPort *port = (VNetPort*)filp->private_data;
   
   start_bh_atomic();
   if (!skb_queue_empty(&port->packetQueue)) {
      end_bh_atomic();
      return 1;
   }
   poll_wait(filp, &port->waitQueue, wait);
   end_bh_atomic();
   return 0;
}

#else

static int
VNetFileOpSelect(struct inode  *inode,
                 struct file   *filp,
                 int            sel_type,
                 select_table  *wait)
{
   VNetPort *port = (VNetPort*)filp->private_data;
   
   switch (sel_type) {
   case SEL_IN:
      start_bh_atomic();
      if (!skb_queue_empty(&port->packetQueue)) {
	 end_bh_atomic();
	 return 1;
      }
      select_wait(&port->waitQueue, wait);
      end_bh_atomic();
      break;
   case SEL_EX:
      // XXX should it be 0 or 1?
      break;
   case SEL_OUT:
      return 1;
   }
   return 0;
}

#endif


/*
 *----------------------------------------------------------------------
 *
 * VNetFileOpIoctl --
 *
 *      The virtual network's ioctl file operation. This is used for
 *      setup of the connection. Currently supported commands are
 *      (taken from sockios.h):
 *
 *      SIOCGIFADDR - get ethernet address          - ioarg OUT: 6 bytes
 *      SIOCSIFADDR - set ethernet address          - ioarg IN:  6 bytes
 *      SIOCGIFFLAGS - get flags                    - ioarg OUT  4 bytes
 *      SIOCSIFFLAGS - set flags                    - ioarg IN:  4 bytes
 *      SIOCGIFBR - get bridging status             - ioarg OUT: 4 bytes
 *      SIOCSIFBR - set bridge peer interface       - ioarg IN:  8 bytes
 *
 *      Private ioctl calls, taken from device-private ioctl space 
 *      in sockios.h, and defined in includes/vm_oui.h:
 *
 *      SIOCSKEEP (0x89F0) - set dealloc flag       - ioarg IN:  1 byte
 *      SIOCGKEEP (0x89F1) - get dealloc flag       - ioarg OUT: 1 byte
 *         By default all hubs (and their associated interfaces) are 
 *         deallocated from memory when their last open port is closed.  
 *         Setting this true "keeps" the hub and thus we will not 
 *         deallocate it until the module is unloaded.
 *
 *      SIOCSLADRF (0x89F2) - set logical address filter (for 
 *         filtering multicast packets)             - ioarg IN:  8 bytes
 *
 *
 *      Supported flags are (taken from if.h):
 *
 *      IFF_UP - ready to receive packets             - OFF by default
 *      IFF_BROADCAST - receive broadcast packets     - OFF by default
 *      IFF_DEBUG - turn on debugging                 - OFF by default
 *      IFF_PROMISC - promiscuous mode                - OFF by default
 *      IFF_MULTICAST - receive multicast packets     - OFF by default
 *      IFF_ALLMULTI - receive all multicast packets 
 *            (like IFF_PROMISC but with multicast)   - OFF by default
 *
 * Results: 
 *      On success 0 else errno.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

int
VNetFileOpIoctl(struct inode   *inode,
                struct file    *filp,
                unsigned int    iocmd,
                unsigned long   ioarg)
{
   VNetPort *port = (VNetPort*)filp->private_data;
   int retval;
   
   switch (iocmd) {
   case SIOCGIFADDR:
      retval = verify_area(VERIFY_WRITE, (void*)ioarg, ETH_ALEN);
      if (retval) {
         return retval;
      }
      memcpy_tofs((void*)ioarg, port->paddr, ETH_ALEN);
      break;
      
   case SIOCSIFADDR:
      retval = verify_area(VERIFY_READ, (void*)ioarg, ETH_ALEN);
      if (retval) {
         return retval;
      }
      memcpy_fromfs(port->paddr, (void*)ioarg, ETH_ALEN);
      if (port->paddr[0] == VMX86_OUI0 &&
          port->paddr[1] == VMX86_OUI1 &&
          port->paddr[2] == VMX86_OUI2 &&
	  (port->paddr[3] & VMX86_MAC_PREFIX) == VMX86_MAC_RANDOM &&
          port->paddr[4] == 0 &&
          port->paddr[5] == 0) {
	 VNetMakeRandomAddress(port->paddr);
      } else if (port->paddr[0] == VMX86_OUI0 &&
		 port->paddr[1] == VMX86_OUI1 &&
		 port->paddr[2] == VMX86_OUI2 &&
		 (port->paddr[3] & VMX86_MAC_PREFIX) == VMX86_MAC_IPBASED) {

	 int i, h, p;
	 Bool ipmacs[VNET_NUM_IPBASED_MACS]; // TRUE == taken, FALSE == available
	 memset(&ipmacs, 0, sizeof ipmacs);
	 ipmacs[0] = TRUE; // our range is 1-63
	 
	 /* search and assign lowest "i" value for IP-based addr.  
	  * Bool[] is inefficient ...  bit vectors?
	  */
	 for (h = 0; h < VNET_NUM_HUBS; h++) {
	    if (!vnetHub[h]) {
	       continue;
	    }
	    for (p = 0; p < VNET_PORTS_PER_HUB; p++) {
	       if ((vnetHub[h]->port[p].paddr[3] & VMX86_MAC_PREFIX) == VMX86_MAC_IPBASED) {
		  int val = (vnetHub[h]->port[p].paddr[3] & 0x3F);  // 3F = 0011 1111

		  if (port == &vnetHub[h]->port[p]) { 
		     ipmacs[val] = FALSE; // current port IS available
		  } else {
		     ipmacs[val] = TRUE;  // available
		  }
	       }
	    }
	 }

	 for (i = 1; i < VNET_NUM_IPBASED_MACS; i++) {
	    if (ipmacs[i] == FALSE) {
	       LOG(0, (KERN_INFO "/dev/vmnet: assigning IP-based address %d.\n", i));
	       port->paddr[3] = VMX86_MAC_IPBASED | i;
	       return 0;
	    }
	 }
	 LOG(0, (KERN_NOTICE "/dev/vmnet: out of IP-based MAC addresses.\n"));
	 return -EBUSY; // limit reached, vlance should catch this
      }
      break;

   case SIOCSLADRF:
      memcpy_fromfs(port->ladrf, (void*)ioarg, 8);
      break;

   case SIOCGKEEP:
      retval = verify_area(VERIFY_WRITE, (void*)ioarg, 1);
      if (retval) {
	 return retval;
      }
      put_user(port->hub->keep, (Bool*)ioarg);
      break;

   case SIOCSKEEP:
      if (ioarg) {
	 port->hub->keep = TRUE;
      } else {
	 port->hub->keep = FALSE;
      }
      break;

   case SIOCGIFFLAGS:
      retval = verify_area(VERIFY_WRITE, (void*)ioarg, 4);
      if (retval) {
         return retval;
      }
      put_user(port->flags, (uint32*)ioarg);
      break;
      
   case SIOCSIFFLAGS:
      port->flags = ((ioarg
                      & (IFF_UP|IFF_BROADCAST|IFF_DEBUG|IFF_PROMISC|IFF_MULTICAST|IFF_ALLMULTI))
                     | IFF_RUNNING);
      /* XXX process queued packets? */
      break;

   case SIOCGIFBR: {
      VNetHub *hub = port->hub;
      int flags = 0;
      int i;
      retval = verify_area(VERIFY_WRITE, (void *)ioarg, sizeof flags);
      if (retval) {
         return retval;
      }
      for (i = 0; i < VNET_PORTS_PER_HUB; i++) {
	 VNetPort *p = &hub->port[i];
	 if (p->bridge != NULL) {
	    flags |= p->bridge->dev == NULL ? 1 : 2;
	 }
      }
      put_user(flags, (int *)ioarg);
      break;
   }

   case SIOCSIFBR:
      retval = VNetBridgeAttach(port, ioarg);
      if (retval) {
	 return retval;
      }
      break;

   default:
      return -ENOIOCTLCMD;
      break;
   }

   return 0;
}


/*
 *----------------------------------------------------------------------
 *
 * VNetNetifOpen --
 *
 *      The virtual network's open dev operation. 
 *
 * Results: 
 *      errno.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

int
VNetNetifOpen(struct device *dev)
{
   /*
    * The host interface is not available if the hub is bridged.
    *
    * It's actually okay to support both.  We just need
    * to tag packets when VNetXmitPacket gives them to the interface
    * so they can be dropped by VNetBridgeReceive().
    */

   if (((VNetHub *)dev->priv)->bridgeCount > 0) {
      return -EBUSY;
   }

   /*
    * set keep on.  if a user's using the dev side of the vnet, he probably
    * want to keep his settings.
    */
   ((VNetHub *)dev->priv)->keep = TRUE;

   dev->tbusy = 0;
   dev->interrupt = 0;
   dev->start = 1;
   MOD_INC_USE_COUNT;
   return 0;
}


/*
 *----------------------------------------------------------------------
 *
 * VNetNetifProbe --
 *
 *      ???
 *
 * Results: 
 *      0.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

int
VNetNetifProbe(struct device *dev)
{
   return 0;
}


/*
 *----------------------------------------------------------------------
 *
 * VNetNetifClose --
 *
 *      The virtual network's close dev operation. 
 *
 * Results: 
 *      errno.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

int
VNetNetifClose(struct device *dev)
{
   dev->start = 0;
   dev->tbusy = 1;
   MOD_DEC_USE_COUNT;
   return 0;
}


/*
 *----------------------------------------------------------------------
 *
 * VNetNetifStartXmit --
 *
 *      The virtual network's start xmit dev operation. 
 *
 * Results: 
 *      ???, 0.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

int
VNetNetifStartXmit(struct sk_buff  *skb,
                   struct device   *dev)
{
   VNetHub *hub = (VNetHub*)dev->priv;

   if(skb == NULL) {
#ifndef KERNEL_2_1
      dev_tint(dev);
#endif
      return 0;
   }

   /* Block a timer-based transmit from overlapping.  This could better be
    * done with atomic_swap(1, dev->tbusy), but set_bit() works as well.
    * If this ever occurs the queue layer is doing something evil!
    */
#ifdef KERNEL_2_1
   if (test_and_set_bit(0, (void*)&dev->tbusy) != 0)
#else
   if (set_bit(0, (void*)&dev->tbusy) != 0)
#endif
   {
      LOG(0, (KERN_NOTICE "%s: transmitter access conflict.\n", dev->name));
      return 1;
   }

   VNetXmitPacket(hub, NULL, skb);

   hub->stats.tx_packets++;
   dev->tbusy = 0;
   dev->trans_start = jiffies;

   return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * VNetNetifSetMAC --
 *
 *      Sets MAC address (i.e. via ifconfig) of netif device.
 *
 * Results: 
 *      Errno.
 *
 * Side effects:
 *      The MAC address may be changed.
 *
 *----------------------------------------------------------------------
 */
int
VNetNetifSetMAC(struct device *dev, void *p)
{
   struct sockaddr *addr = p;

   /* no check, it's okay to change while interface is "up."
       if (dev->start) { 
       return -EBUSY;
       } 
   */
   memcpy(dev->dev_addr, addr->sa_data, dev->addr_len);
   return 0; // XXX should check for 00:50:56:XX:XX:XX?
}

/*
 *----------------------------------------------------------------------
 *
 * VNetNetifSetMulticast --
 *
 *      Sets or clears the multicast address list.  This information
 *      comes from an array in dev->mc_list, and with a counter in
 *      dev->mc_count.
 *
 *      Since host-only network ifaces can't be bridged, it's debatable
 *      whether this is at all useful, but at least now you can turn it 
 *      on from ifconfig without getting an ioctl error.
 * Results: 
 *      Void.
 *
 * Side effects:
 *      Multicast address list might get changed.
 *
 *----------------------------------------------------------------------
 */
void
VNetNetifSetMulticast(struct device *dev)
{

}

/*
 *----------------------------------------------------------------------
 *
 * VNetNetifGetStats --
 *
 *      The virtual network's get stats dev operation. 
 *
 * Results: 
 *      A struct full of stats.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

struct enet_statistics *
VNetNetifGetStats(struct device *dev)
{
   VNetHub *hub = (VNetHub*)dev->priv;
   return &(hub->stats);
}


/*
 *----------------------------------------------------------------------
 *
 * VNetBridgeAttach --
 *
 *      Attach a port to a peer interface
 *
 * Results: 
 *      errno.
 *
 * Side effects:
 *      Some of the bridging state is set up.
 *	If the interface is not there, we'll hang around and
 *	try to bring it up it in VNetBridgeNotify.
 *
 *----------------------------------------------------------------------
 */

int
VNetBridgeAttach(VNetPort *port, unsigned long ioarg)
{
   VNetHub *hub = port->hub;
   VNetBridge *bridge = NULL;
   char peerName[sizeof bridge->name];
   int retval = 0;

   /*
    * Get peer device name
    */

   retval = verify_area(VERIFY_READ, (void *)ioarg, sizeof peerName);
   if (retval != 0) {
      goto out;
   }
   memcpy_fromfs(peerName, (void *)ioarg, sizeof peerName);
   peerName[sizeof peerName - 1] = 0;

   /*
    * If device name is empty, detach the bridge.
    */

   if (peerName[0] == '\0') {
      if (port->bridge == NULL) {
	 retval = -EINVAL;
      } else {
	 VNetBridgeDetach(port);
      }
      goto out;
   }

   /*
    * It's an error to be already bridging.
    */

   if (port->bridge != NULL) {
      retval = -EBUSY;
      goto out;
   }

   /*
    * Allocate bridge structure
    */

   bridge = kmalloc(sizeof *bridge, GFP_USER);
   if (bridge == NULL) {
      retval = -ENOMEM;
      goto out;
   }
   memset(bridge, 0, sizeof *bridge);
   memcpy(bridge->name, peerName, sizeof bridge->name);
   bridge->port = port;
   port->bridge = bridge;

   /*
    * Set up notifier for network device state change
    */

   bridge->notifier.notifier_call = VNetBridgeNotify;
   bridge->notifier.priority = 0;
   register_netdevice_notifier(&bridge->notifier);

   /*
    * Try to bring it up
    */

   retval = VNetBridgeUp(port);
   if (retval == -ENODEV) {
      LOG(1, (KERN_DEBUG "%s: bridge peer interface %s not found, "
	      "will wait for it to come up\n",
	      hub->devName, bridge->name));
      retval = 0;
   }
   if (retval != 0) {
      goto out;
   }

   /*
    * Return
    */

   LOG(1, (KERN_DEBUG "%s: bridge %s attached\n", hub->devName, bridge->name));
out:
   if (retval != 0) {

      /*
       * The following check must be on bridge, not port->bridge,
       * because we also come here with an existing bridge,
       * which we must not deallocate here.
       */

      if (bridge != NULL) {
	 if (bridge->notifier.notifier_call != NULL) {
	    unregister_netdevice_notifier(&bridge->notifier);
	 }
	 kfree(bridge);
	 port->bridge = NULL;
      }
   }
   return retval;
}


/*
 *----------------------------------------------------------------------
 *
 * VNetBridgeUp --
 *
 *      Bring a bridge up
 *
 * Results: 
 *      errno.
 *
 * Side effects:
 *      Bridging may be brought up with a peer interface.
 *
 *----------------------------------------------------------------------
 */

int
VNetBridgeUp(VNetPort *port)
{
   VNetHub *hub = port->hub;
   VNetBridge *bridge = port->bridge;
   int i;
   int retval = 0;

   if (bridge->dev != NULL) {
      LOG(0, (KERN_NOTICE "%s: bridge %s already up\n", hub->devName, bridge->name));
      goto out;
   }

   /*
    * Get peer device structure
    */

   DEV_LOCK_LIST();
   bridge->dev = dev_get(bridge->name);
   LOG(2, (KERN_DEBUG "%s: bridge %s got dev 0x%08lx\n",
	   hub->devName, bridge->name, (uint32) bridge->dev));
   if (bridge->dev == NULL) {
      DEV_UNLOCK_LIST();
      retval = -ENODEV;
      goto out;
   }

   /*
    * At a minimum, the header size should be the same as ours.
    *
    * XXX we should either do header translation or ensure this
    * is an Ethernet.
    */

   if (bridge->dev->hard_header_len != ETH_HLEN) {
      LOG(1, (KERN_DEBUG "%s: can't bridge with %s, bad header length %d\n",
	      hub->devName, bridge->name, bridge->dev->hard_header_len));
      DEV_UNLOCK_LIST();
      retval = -EINVAL;
      goto out;
   }

   /*
    * Are we already bridging the same device?
    */

   for (i = 0; i < VNET_PORTS_PER_HUB; i++) {
      VNetPort *p = &hub->port[i];
      if (p != port && p->bridge != NULL && p->bridge->dev == bridge->dev) {
	 DEV_UNLOCK_LIST();
	 retval = -EBUSY;
	 goto out;
      }
   }

   /*
    * Is the interface up?
    */

   if (hub->netIF->start) {
      DEV_UNLOCK_LIST();
      retval = -EISCONN;	// not a good error message
      goto out;
   }

   /*
    * Are we bridging ourselves?
    *
    * See also comments in VNetNetifOpen.
    */

   if (bridge->dev == hub->netIF) {
      DEV_UNLOCK_LIST();
      retval = -EINVAL;
      goto out;
   }

   /*
    * Get a socket to play with
    *
    * We set the dead field so we don't get a call back from kfree_skb().
    * (The alternative is to support the callback.)
    */

   bridge->sk = SK_ALLOC(GFP_USER);
   if (bridge->sk == NULL) {
      DEV_UNLOCK_LIST();
      retval = -ENOMEM;
      goto out;
   }
   bridge->sk->dead = 1;
   *(VNetPort **)&bridge->sk->protinfo = port;

   /*
    * Link up with the peer device
    */

   bridge->pt.func = VNetBridgeReceive;
   bridge->pt.type = htons(ETH_P_ALL);
   bridge->pt.data = bridge->sk;
   bridge->pt.dev = bridge->dev;
   bridge->promiscSaved = FALSE;
   dev_add_pack(&bridge->pt);

   /*
    * Finish up
    */

   DEV_UNLOCK_LIST();
   LOG(1, (KERN_DEBUG "%s: bridge %s up\n", hub->devName, bridge->name));

   /*
    * Update bridging state
    */

   hub->bridgeCount++;
   if (hub->portCount > 1) {
      VNetBridgeStartPromisc(port);
   }

   /*
    * Return
    */

out:
   if (retval != 0) {
      if (bridge->sk != NULL) {
	 sk_free(bridge->sk);
	 bridge->sk = NULL;
      }
      bridge->dev = NULL;
   }
   return retval;
}


/*
 *----------------------------------------------------------------------
 *
 * VNetBridgeDetach --
 *
 *      Detach a bridged peer interface
 *
 * Results: 
 *      None.
 *
 * Side effects:
 *      Bridging is removed.
 *
 *----------------------------------------------------------------------
 */

void
VNetBridgeDetach(VNetPort *port)
{
   VNetHub *hub = port->hub;
   VNetBridge *bridge = port->bridge;

   if (bridge->dev != NULL) {
      VNetBridgeDown(port);
   }

   port->bridge = NULL;
   unregister_netdevice_notifier(&bridge->notifier);
   LOG(1, (KERN_DEBUG "%s: bridge %s detached\n", hub->devName, bridge->name));
   kfree(bridge);
}


/*
 *----------------------------------------------------------------------
 *
 * VNetBridgeDown --
 *
 *      Bring a bridge down
 *
 * Results: 
 *      None.
 *
 * Side effects:
 *      Bridging is brought down.
 *
 *----------------------------------------------------------------------
 */

void
VNetBridgeDown(VNetPort *port)
{
   VNetHub *hub = port->hub;
   VNetBridge *bridge = port->bridge;

   if (bridge->dev == NULL) {
      LOG(0, (KERN_NOTICE "%s: bridge %s already down\n", hub->devName, bridge->name));
      return;
   }

   hub->bridgeCount--;
   if (hub->portCount > 1) {
      VNetBridgeStopPromisc(port);
   }
   bridge->dev = NULL;
   dev_remove_pack(&bridge->pt);
   sk_free(bridge->sk);
   bridge->sk = NULL;
   LOG(1, (KERN_DEBUG "%s: bridge %s down\n", hub->devName, bridge->name));
}


/*
 *----------------------------------------------------------------------
 *
 * VNetBridgeNotify --
 *
 *      Callback on peer device state change
 *
 * Results: 
 *      NOTIFY_DONE
 *
 * Side effects:
 *      Bridging is disabled or enabled as the case may be.
 *
 *----------------------------------------------------------------------
 */

int
VNetBridgeNotify(struct notifier_block *this, u_long msg, void *data)
{
   VNetBridge *bridge = (VNetBridge *) this;
   VNetPort *port = bridge->port;
   VNetHub *hub = port->hub;
   struct device *dev = (struct device *) data;

   LOG(2, (KERN_DEBUG "%s: bridge notify 0x%lx (%s), dev 0x%p (%s)\n",
	   hub->devName,
	   msg,
	   msg == NETDEV_DOWN ? "NETDEV_DOWN" :
	   msg == NETDEV_UP ? "NETDEV_UP" : "???",
	   dev, dev->name));

   switch (msg) {
   case NETDEV_DOWN:
      if (dev == bridge->dev) {
	 LOG(1, (KERN_DEBUG "%s: lost bridge peer %s\n", hub->devName, bridge->name));
	 VNetBridgeDown(port);
      }
      break;

   case NETDEV_UP:
      if (bridge->dev == NULL && strcmp(dev->name, bridge->name) == 0) {
	 int errno;
	 LOG(1, (KERN_DEBUG "%s: found bridge peer %s\n", hub->devName, bridge->name));
	 errno = VNetBridgeUp(port);
	 switch (-errno) {
	 case 0:
	    break;
	 case ENODEV:
	    LOG(0, (KERN_WARNING "%s: bridge peer interface %s not found",
		    hub->devName, bridge->name));
	    break;
	 case EINVAL:
	    LOG(0, (KERN_WARNING "%s: bridge peer %s is not a valid Ethernet interface\n",
		    hub->devName, bridge->name));
	    break;
	 case EBUSY:
	    LOG(0, (KERN_WARNING "%s: already bridged with %s\n",
		    hub->devName, bridge->name));
	    break;
	 case EISCONN:
	    LOG(0, (KERN_WARNING "%s: already attached as a network interface, "
		    "can't bridge with %s\n",
		    hub->devName, bridge->name));
	    break;
	 case ENOMEM:
	    LOG(0, (KERN_WARNING "%s: bridge %s failed to allocate memory\n",
		    hub->devName, bridge->name));
	    break;
	 default:
	    LOG(0, (KERN_WARNING "%s: failed to bring up bridge %s (error %d)\n",
		    hub->devName, bridge->name, -errno));
	 }
      }
      break;
   }

   return NOTIFY_DONE;
}

/*
 *----------------------------------------------------------------------
 *
 * VNetBridgeReceive --
 *
 *      Receive a packet from a bridged peer device
 *
 *      This is called from the bottom half.  Must be careful.
 *
 * Results: 
 *      errno.
 *
 * Side effects:
 *      A packet may be sent to the vnet.
 *
 *----------------------------------------------------------------------
 */

int
VNetBridgeReceive(struct sk_buff *skb,
		  struct device *dev,
		  struct packet_type *pt)
{
   VNetPort *port = *(VNetPort **)&((struct sock *)pt->data)->protinfo;
   VNetBridge *bridge = port->bridge;
   int i;

   if (bridge == NULL || bridge->dev == NULL) {
      LOG(3, (KERN_DEBUG "%s: bridge %s received %d closed\n",
	      port->hub->devName, dev->name, (int) skb->len));
      DEV_KFREE_SKB(skb, FREE_READ);
      return -EIO;	// value is ignored anyway
   }

   // XXX need to lock history
   for (i = 0; i < VNET_BRIDGE_HISTORY; i++) {
      struct sk_buff *s = bridge->history[i];
      if (s != NULL &&
	  (s == skb || SKB_IS_CLONE_OF(skb, s))) {
	 bridge->history[i] = NULL;
	 KFREE_SKB(s, FREE_WRITE);
	 LOG(3, (KERN_DEBUG "%s: bridge %s receive %d self %d\n",
		 port->hub->devName, dev->name, (int) skb->len, i));
	 // FREE_WRITE because we did the allocation, it's not used anyway
	 DEV_KFREE_SKB(skb, FREE_WRITE);
	 return 0;
      }
   }

#  if LOGLEVEL >= 4
   {
      struct timeval now;
      do_gettimeofday(&now);
      LOG(3, (KERN_DEBUG "%s: bridge %s time %d.%06d\n",
	      port->hub->devName, bridge->name,
	      now.tv_sec - vnetTime.tv_sec, now.tv_usec - vnetTime.tv_usec));
   }
#  endif

   skb_push(skb, skb->data - skb->mac.raw);
   LOG(3, (KERN_DEBUG "%s: bridge %s receive %d\n",
	   port->hub->devName, bridge->name, (int) skb->len));

#  ifdef EDWARD
   VNetPacketDebug(skb);
#  endif

   // XXX shouldn't send to netIF?
   VNetXmitPacket(port->hub, port, skb);

   return 0;
}


/*
 *----------------------------------------------------------------------
 *
 * VNetBridgeXmitPacket --
 *
 *      Transmit a packet to a bridged peer device
 *
 * Results: 
 *      None.
 *
 * Side effects:
 *      A packet is sent to the device and to the host via the device.
 *
 *----------------------------------------------------------------------
 */

void
VNetBridgeXmitPacket(VNetPort *port, struct sk_buff *skb)
{
   struct device *dev = port->bridge->dev;
   struct VNetBridge *bridge = port->bridge;
   uint8 *dest = SKB_2_DESTMAC(skb);
   int i;
   struct sk_buff *clone;

   LOG(3, (KERN_DEBUG "%s: bridge %s transmit %d\n",
	   port->hub->devName, bridge->name, (int) skb->len));

   /*
    * Check promiscuous bit periodically
    */

   VNetBridgeCheckPromisc(port);

   /*
    * We need to send the packet both up to the host and down
    * to the interface.
    * However, we ignore packets destined only for this hub.
    */

   for (i = 0; i < VNET_PORTS_PER_HUB; i++) {
      VNetPort *p = &port->hub->port[i];
      if (UP_AND_RUNNING(p->flags) && MAC_EQ(dest, p->paddr)) {
	 return;
      }
   }

   /*
    * Send down (imitate packet_sendmsg)
    *
    * Do this only if the packet is not addressed to the peer,
    * and the packet size is not too big.
    */

   DEV_LOCK_LIST();
   if (MAC_EQ(dest, dev->dev_addr) ||
       skb->len > dev->mtu + dev->hard_header_len) {
      DEV_UNLOCK_LIST();
   } else {
#     if 0 // XXX we should do header translation
      if ((dev->flags & IFF_SOFTHEADERS) != 0) {
	 if (skb->len > dev->mtu) {
	    clone = NULL;
	 } else {
	    clone = alloc_skb(skb->len + dev->hard_header_len, GFP_ATOMIC);
	 }
	 if (clone != NULL) {
	    skb_reserve(clone, dev->hard_header_len);
	    if (dev->hard_header != NULL) {
	       dev->hard_header(clone, dev, ETH_P_IP, NULL, NULL, skb->len);
	    }
	    memcpy(skb_put(clone, skb->len), skb->data, skb->len);
	 }
      }
#     endif
      clone = skb_clone(skb, GFP_ATOMIC);
      if (clone == NULL) {
	 DEV_UNLOCK_LIST();
      } else {
	 struct sock *sk = bridge->sk;
	 atomic_add(skb->truesize, &sk->wmem_alloc);
#	 ifndef KERNEL_2_1
	 clone->free = 1;
	 clone->arp = 1;
#	 endif
	 clone->sk = sk;
	 clone->protocol = ((struct ethhdr *)skb->data)->h_proto; // XXX
	 if ((dev->flags & IFF_UP) != 0) {
	    DEV_UNLOCK_LIST();
	    DEV_QUEUE_XMIT(clone, dev, 0);
	 } else {
	    DEV_UNLOCK_LIST();
	    KFREE_SKB(clone, FREE_WRITE);
	 }
      }
   }

   /*
    * Send up (imitate Ethernet receive)
    *
    * Do this if the packet is addressed to the peer (or is broadcast, etc.).
    *
    * This packet will get back to us, via VNetBridgeReceive.
    * We save it so we can recognize it (and its clones) again.
    *
    * netif_rx is an interrupt-time function.  Wrap it in cli, sti.
    */

   if (VNetPacketMatch(dest, dev->dev_addr, (uint8 *)&AllMultiFilter, dev->flags)) {
      clone = skb_clone(skb, GFP_ATOMIC);
      if (clone) {
	 unsigned long flags;
	 int i;
	 SKB_INCREF(clone);
	 // XXX need to lock history
	 for (i = 0; i < VNET_BRIDGE_HISTORY; i++) {
	    if (bridge->history[i] == NULL) {
	       bridge->history[i] = clone;
#	       if LOGLEVEL >= 3
	       {
		  int j;
		  int count = 0;
		  for (j = 0; j < VNET_BRIDGE_HISTORY; j++) {
		     if (bridge->history[j] != NULL) {
			count++;
		     }
		  }
		  LOG(3, (KERN_DEBUG "%s: bridge %s host slot %d history %d\n",
			  port->hub->devName, bridge->name, i, count));
	       }
#	       endif
	       break;
	    }
	 }
	 if (i >= VNET_BRIDGE_HISTORY) {
	    LOG(1, (KERN_NOTICE "%s: bridge %s history full\n",
		    port->hub->devName, bridge->name));
	    for (i = 0; i < VNET_BRIDGE_HISTORY; i++) {
	       struct sk_buff *s = bridge->history[i];
	       bridge->history[i] = NULL;
	       KFREE_SKB(s, FREE_WRITE);
	    }
	    bridge->history[0] = clone;
	 }

#	 ifdef EDWARD
	 VNetPacketDebug(clone);
#	 endif

	 clone->dev = dev;
	 clone->protocol = eth_type_trans(clone, dev);
	 save_flags(flags);
	 cli();
	 netif_rx(clone);
#	 if LOGLEVEL >= 4
	 do_gettimeofday(&vnetTime);
#	 endif
	 restore_flags(flags);
      }
   }
}


/*
 *----------------------------------------------------------------------
 *
 * VNetBridgeStartPromisc --
 *
 *      Set IFF_PROMISC on the peer interface
 *
 *      XXX need locking?
 *
 * Results: 
 *      None.
 *
 * Side effects:
 *      The peer interface IFF_PROMISC flag may be changed.
 *
 *----------------------------------------------------------------------
 */

void
VNetBridgeStartPromisc(VNetPort *port)
{
   VNetBridge *bridge = port->bridge;
   struct device *dev = bridge->dev;
   Bool devPromisc = (dev->flags & IFF_PROMISC) != 0;

   if (!bridge->promiscSaved) {
      bridge->promiscSaved = TRUE;
      bridge->savedPromisc = devPromisc;
   }
   if (!devPromisc) {
      LOG(1, (KERN_DEBUG "%s: bridge %s set IFF_PROMISC\n",
	      port->hub->devName, bridge->name));
      dev->flags |= IFF_PROMISC;
      if (dev->set_multicast_list != NULL) {
	 dev->set_multicast_list(dev);
      }
   }
}


/*
 *----------------------------------------------------------------------
 *
 * VNetBridgeStopPromisc --
 *
 *      Restore saved IFF_PROMISC on the peer interface
 *
 *      XXX need locking?
 *
 * Results: 
 *      None.
 *
 * Side effects:
 *      The peer interface IFF_PROMISC flag may be changed.
 *
 *----------------------------------------------------------------------
 */

void
VNetBridgeStopPromisc(VNetPort *port)
{
   VNetBridge *bridge = port->bridge;
   struct device *dev = bridge->dev;
   Bool devPromisc = (dev->flags & IFF_PROMISC) != 0;

   /*
    * If somebody else has cleared the bit on us, then that's
    * what they want, and we don't touch it.
    */

   if (devPromisc && bridge->promiscSaved && !bridge->savedPromisc) {
      LOG(1, (KERN_DEBUG "%s: bridge %s clear IFF_PROMISC\n",
	      port->hub->devName, bridge->name));
      dev->flags &= ~IFF_PROMISC;
      if (dev->set_multicast_list != NULL) {
	 dev->set_multicast_list(dev);
      }
   }
   bridge->promiscSaved = FALSE;
}


/*
 *----------------------------------------------------------------------
 *
 * VNetBridgeCheckPromisc --
 *
 *      Make sure IFF_PROMISC on the peer interface is set
 *
 *      This can be called periodically.
 *
 *      XXX need locking?
 *
 * Results: 
 *      None.
 *
 * Side effects:
 *      The peer interface IFF_PROMISC flag may be changed.
 *
 *----------------------------------------------------------------------
 */

void
VNetBridgeCheckPromisc(VNetPort *port)
{
   VNetBridge *bridge = port->bridge;
   struct device *dev = bridge->dev;
   Bool devPromisc = (dev->flags & IFF_PROMISC) != 0;

   if (bridge->promiscSaved && !devPromisc) {
      LOG(1, (KERN_DEBUG "%s: bridge %s lost IFF_PROMISC\n",
	      port->hub->devName, bridge->name));
      bridge->savedPromisc = devPromisc;
      dev->flags |= IFF_PROMISC;
      if (dev->set_multicast_list != NULL) {
	 dev->set_multicast_list(dev);
      }
   }
}


/*
 *----------------------------------------------------------------------
 *
 * VNetXmitPacket --
 *
 *      Transmit the packet to all the ports on the virtual network,
 *      except the source port. If the source is NULL then the source is
 *      the host network interface.
 *
 * Results: 
 *      None.
 *
 * Side effects:
 *      Gives up ownership of skb.
 *
 *----------------------------------------------------------------------
 */

void
VNetXmitPacket(VNetHub         *hub,
               VNetPort        *src, 
               struct sk_buff  *skb)
{
   VNetPort *port;
   struct sk_buff *clone;
   uint8 *dest = SKB_2_DESTMAC(skb);
   int i;

   /*
    * Send packet to each port
    */

   for (i = 0; i < VNET_PORTS_PER_HUB; i++) {
      port = &hub->port[i];

      /*
       * Is this packet for us?
       */

      if ((port == src) || !UP_AND_RUNNING(port->flags)) {
         continue;
      }
      if (!VNetPacketMatch(dest, port->paddr, port->ladrf, port->flags)) {
	 continue;
      }

      /*
       * If this port is a bridge, then send the packet to the bridged
       * interface only.
       */

      if (port->bridge != NULL && port->bridge->dev != NULL) {
	 VNetBridgeXmitPacket(port, skb);
	 continue;
      }

      /*
       * Ignore if port queue is full.
       */

      if (skb_queue_len(&port->packetQueue) >= VNET_MAX_QLEN) {
         continue;
      }

      /*
       * Queue the packet for this port
       *
       * We can be called from the backend (via VNetBridgeReceive),
       * so we need to lock around skb queue manipulation.
       *
       * XXX are we missing any other places?
       */

      clone = skb_clone(skb, GFP_ATOMIC);
      if (clone) {
	 start_bh_atomic();
	 skb_queue_tail(&port->packetQueue, clone);
	 end_bh_atomic();
	 wake_up(&port->waitQueue);
      }
   }

   if (src && UP_AND_RUNNING(hub->netIF->flags)
       && VNetPacketMatch(dest, hub->netIF->dev_addr, (uint8 *)AllMultiFilter, 
			  hub->netIF->flags)) {
      /* send to the host interface */
      skb->protocol = eth_type_trans(skb, hub->netIF);
      netif_rx(skb);
      hub->stats.rx_packets++;
   } else {
      DEV_KFREE_SKB(skb, FREE_WRITE);
   }
}

/*
 *----------------------------------------------------------------------
 *
 * VNetMulticastFilter --
 *
 *      Utility function that filters multicast packets according
 *      to a 64-bit logical address filter (like the one on the
 *      lance chipset).  AllMultiFilter lets all packets through.
 *
 *      We generate a hash value from the destination MAC address
 *      and see if it's in our filter.  Broadcast packets have
 *      already OK'd by PacketMatch, so we don't have to worry
 *      about that.
 *
 *      (This is in the green AMD "Ethernet Controllers" book, 
 *      page 1-53.)
 *
 * Results:
 *      TRUE if packet is in filter, FALSE if not.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

#define CRC_POLYNOMIAL_BE 0x04c11db7UL  /* Ethernet CRC, big endian */

Bool
VNetMulticastFilter(uint8 *destAddr,
		    uint8 *ladrf)
{
   uint16 hashcode;
   int32 crc, poly = CRC_POLYNOMIAL_BE;
   int j, bit, byte;
   
   crc = 0xffffffff;                  /* init CRC for each address */
   for (byte=0; byte<6; byte++) {     /* for each address byte */
      /* process each address bit */ 
      for (bit = *destAddr++,j=0;j<8;j++, bit>>=1) {
	 crc = (crc << 1) ^ ((((crc<0?1:0) ^ bit) & 0x01) ? poly : 0);
      }
   }
   hashcode = (crc & 1);              /* hashcode is 6 LSb of CRC ... */
   for (j=0; j<5; j++) {                /* ... in reverse order. */
      hashcode = (hashcode << 1) | ((crc>>=1) & 1);
   }                                      
   
   byte = hashcode >> 3;              /* bit[3-5] -> byte in filter */
   bit = 1 << (hashcode & 0x07);      /* bit[0-2] -> bit in byte */
   if (ladrf[byte] & bit) {
      return TRUE;
   } else {
      return FALSE;
   }
}

/*
 *----------------------------------------------------------------------
 *
 * VNetPacketMatch --
 *
 *      Determines whether the packet should be given to the interface.
 *
 * Results: 
 *      TRUE if the pasket is OK for this interface, FALSE otherwise.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

Bool
VNetPacketMatch(uint8   *destAddr,
                uint8   *ifAddr,
		uint8   *ladrf,
                uint32   flags)
{
   static uint8 broadcast[] = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff};
   return ((flags & IFF_PROMISC)
           || MAC_EQ(destAddr, ifAddr)
           || ((flags & IFF_BROADCAST) && MAC_EQ(destAddr, broadcast))
           || ((destAddr[0] & 0x1) && 
	       (flags & IFF_ALLMULTI || // allow all multi
		     (flags & IFF_MULTICAST && 
		      VNetMulticastFilter(destAddr, ladrf)))));
}

#ifdef i386

static inline unsigned long long GET_TSC(void)
{
    unsigned long long tim;
   __asm__ volatile (".byte 0x0f,0x31"
           :"=A" (tim));
    return tim;
}

#endif 

/*
 *----------------------------------------------------------------------
 *
 * VNetMakeRandomAddress --
 *
 *      Make the low 22 bits of a MAC address random.
 *	The address is chosen to be unique on this host.
 *
 * Results: 
 *      None.
 *
 * Side effects:
 *      The address is changed.
 *
 *----------------------------------------------------------------------
 */

void
VNetMakeRandomAddress(uint8 *addr)
{
   Bool conflict;

   do {
      int i, j;
      uint32 r = GET_TSC();
      addr[3] = VMX86_MAC_RANDOM | (r >> 16 & ~VMX86_MAC_PREFIX);
      addr[4] = r >> 8;
      addr[5] = r;

      conflict = FALSE;
      for (i = 0; i < VNET_NUM_HUBS && !conflict; i++) {
	 VNetHub *hub = vnetHub[i];

	 if (!hub) {     // hub may not have been alloc'd yet.
	     continue;
	 }
	 if (hub->netIF->dev_addr != addr &&
	     MAC_EQ(hub->netIF->dev_addr, addr)) {
	    conflict = TRUE;
	    break;
	 }
	 for (j = 0; j < VNET_PORTS_PER_HUB; j++) {
	    VNetPort *port = &hub->port[j];
	    if ((port->flags & IFF_RUNNING) != 0 &&
		port->paddr != addr &&
		MAC_EQ(port->paddr, addr)) {
	       conflict = TRUE;
	       break;
	    }
	 }
      }
   } while (conflict);
}


#ifdef EDWARD
#include <linux/ip.h>
#include <linux/icmp.h>

void
VNetPacketDebug(struct sk_buff *skb)
{
   char *packet = skb->data;
   struct ethhdr *eh = (struct ethhdr *)packet;
   struct iphdr *iph = (struct iphdr *)(packet + sizeof(struct ethhdr));
   struct icmphdr *icmph = (struct icmphdr *)(packet + sizeof (struct ethhdr) +
						       sizeof (struct iphdr));
   if (ntohs(eh->h_proto) == ETH_P_IP &&
       iph->protocol == IPPROTO_ICMP &&
       (icmph->type == ICMP_ECHO || icmph->type == ICMP_ECHOREPLY)) {
      int len = ntohs(iph->tot_len) + sizeof (struct ethhdr);
      struct timeval *t =
	 (struct timeval *)(packet + sizeof (struct ethhdr) +
			    sizeof (struct iphdr) +
			    sizeof (struct icmphdr));
      struct timeval *end = (struct timeval *)&packet[len] - 1;
      for (; t <= end; t++) {
	 if (t->tv_sec == 0) {
	    do_gettimeofday(t);
	    icmph->checksum = 0;
	    icmph->checksum =
	       VNetICMPChecksum((uint16 *)icmph, packet + len - (char *)icmph);
	    break;
	 }
      }
   }
}


uint16
VNetICMPChecksum(uint16 *packet, int len)
{
   int n;
   uint16 *p;
   uint32 cs;

   p = packet;
   cs = 0;
   for (n = len; n > 1; n -= 2) {
      cs += *p++;
   }
   if (n > 0) {
      uint8 x[2];
      x[0] = *(uint8 *)p;
      x[1] = 0;
      cs += *(uint16 *)x;
   }
   cs = (cs >> 16) + (cs & 0xffff);
   cs += cs >> 16;
   return ~cs;
}
#endif // EDWARD

/* 
 * VNetProcFileRead: print to /proc/vmnet.  ifdefs for 2.0 and 2.2.
 */

int 
VNetProcFileRead(char *buffer, 
		 char **start, 
		 off_t offset, 
		 int length, 
		 int dummy)
{
   int i, j;
   uint fd;

   off_t begin = 0;
   off_t pos = 0;
   int len = 0;
   int size;

   struct task_struct *p;
   
   size = sprintf(buffer,  
		  "Version: 1.2\n" // so parsers will know 
               //   123456 12345678 1234 1234 12345678901234567 12 1234567 12345 12345 12345678 123
		  "    PID Hub      Port Keep MAC ethernet addr Up Promisc Bcast Mcast AllMulti Dbg\n"
		  "------- -------- ---- ---- ----------------- -- ------- ----- ----- -------- ---\n");
   pos+=size; 
   len+=size;

/*
   for (i = 0; i < VNET_NUM_HUBS; i++) {  // a little extra
      if (!vnetHub[i]) {
	 continue;
      }
      if (vnetHub[i]->keep) {
	 len += sprintf(buffer+len, "Preserving configuration on interface/hub vmnet%d\n", i);
      }
   }
*/   
   for (p = current->next_task; p != current; p = p->next_task) {
	
#ifdef KERNEL_2_1
      for (fd = 0; p->files && fd < p->files->max_fds; fd++) {
	 struct file * file = fcheck_task(p,fd);
	 if (!file || !file->f_dentry)
	    continue;
	 j = MAJOR(file->f_dentry->d_inode->i_rdev); 
	 if (j == VNET_MAJOR_NUMBER) {
	    VNetPort *port = file->private_data;
#else
      for (fd = 0; p->files && fd < NR_OPEN; fd++) {
	 if (!p->files)
	    break;
	 if (!p->files->fd[fd] || !p->files->fd[fd]->f_inode)
	    continue;
	 j = MAJOR(p->files->fd[fd]->f_inode->i_rdev); 
	 if (j == VNET_MAJOR_NUMBER) {
	    VNetPort *port = p->files->fd[fd]->private_data;
#endif
	    i = (port - port->hub->port);
	    size = sprintf(buffer+len, 
			   "%7d %-8s %4d %4s %02x:%02x:%02x:%02x:%02x:%02x %2s %7s %5s %5s %8s %3s\n", 
			   p->pid, 
			   port->hub->devName, 
			   i,  
			   port->hub->keep ? "Y" : "N",
			   port->paddr[0],
			   port->paddr[1],
			   port->paddr[2],
			   port->paddr[3],
			   port->paddr[4],
			   port->paddr[5],
			   port->flags & IFF_UP        ? "Y" : "N",
			   port->flags & IFF_PROMISC   ? "Y" : "N",
			   port->flags & IFF_BROADCAST ? "Y" : "N",
			   port->flags & IFF_MULTICAST ? "Y" : "N",
			   port->flags & IFF_ALLMULTI  ? "Y" : "N",
			   port->flags & IFF_DEBUG     ? "Y" : "N");
	 } else {
	    size = 0;
	 }

	 // buffer control
	 len+=size;
	 pos=begin+len;
	 
	 if(pos<offset) {
	    len=0;
	    begin=pos;
	 }
	 if(pos>offset+length) {
	    goto out;
	 }
      }
   }
 out:
   *start=buffer+(offset-begin); /* Start of wanted data */
   len-=(offset-begin);		 /* Start slop */
   if(len>length)
      len=length;		 /* Ending slop */
   return len;
}
